package main

import (
	"context"
	"fmt"
	"net"
	"strconv"
	"sync"
	"time"

	"github.com/golang/glog"
	"google.golang.org/grpc"
	pluginapi "k8s.io/kubelet/pkg/apis/deviceplugin/v1beta1"
)

const (
	mpsControlBin      = "/usr/local/nvidia/bin/nvidia-cuda-mps-control"
	mpsActiveThreadCmd = "get_default_active_thread_percentage"
	mpsMemLimitEnv     = "CUDA_MPS_PINNED_DEVICE_MEM_LIMIT"
	mpsThreadLimitEnv  = "CUDA_MPS_ACTIVE_THREAD_PERCENTAGE"
	pluginEndpoint     = "/device-plugin/mps-device-plugin.sock"
	kubeletEndpoint    = "/device-plugin/kubelet.sock"
)

var (
	computeResourceName = "nvidia-mps.com/vcore"
	memResourceName     = "nvidia-mps.com/vmemory"
)

type mpsGPUManager struct {
	grpcServer           *grpc.Server
	devices              map[string]pluginapi.Device
	computePartitionSize int // In utilization (eg. 1%)
	memPartitionSize     int // In MB (eg. 256MB)
	socket               string
	stop                 chan bool
}

func NewMpsGPUManager(computePartitionSize, memPartitionSize int) *mpsGPUManager {
	return &mpsGPUManager{
		devices:              make(map[string]pluginapi.Device),
		computePartitionSize: computePartitionSize,
		memPartitionSize:     memPartitionSize,
		stop:                 make(chan bool),
	}
}

type pluginService struct {
	mgm *mpsGPUManager
}

// func (mgm *mpsGPUManager) isMpsHealthy() error {
// 	glog.Infof("Checking status of MPS on the node")

// 	var out bytes.Buffer
// 	reader, writer := io.Pipe()
// 	defer writer.Close()
// 	defer reader.Close()

// 	mpsCmd := exec.Command(mpsControlBin)
// 	mpsCmd.Stdin = reader
// 	mpsCmd.Stdout = &out

// 	err := mpsCmd.Start()
// 	if err != nil {
// 		return fmt.Errorf("failed to start NVIDIA MPS health check: %v", err)
// 	}

// 	writer.Write([]byte(mpsActiveThreadCmd))
// 	writer.Close()

// 	err = mpsCmd.Wait()
// 	if err != nil {
// 		return fmt.Errorf("failed to health check NVIDIA MPS: %v", err)
// 	}

// 	reader.Close()
// 	glog.Infof("MPS is healthy, active thread percentage = %s", out.String())
// 	return nil
// }

func (mgm *mpsGPUManager) Serve() {
	glog.Infof("Starting MPS GPU Manager")
	lis, err := net.Listen("unix", pluginEndpoint)
	if err != nil {
		glog.Fatal("starting device plugin server failed : %v", err)
	}
	mgm.socket = pluginEndpoint
	mgm.grpcServer = grpc.NewServer()

	pluginbeta := &pluginService{mgm: mgm}
	pluginbeta.RegisterService()
	registeredWithKubelet := false

	for {
		select {
		case <-mgm.stop:
			close(mgm.stop)
			return
		default:
			{
				var wg sync.WaitGroup
				wg.Add(1)
				// Starts device plugin service.
				go func() {
					defer wg.Done()
					// Blocking call to accept incoming connections.
					err := mgm.grpcServer.Serve(lis)
					glog.Errorf("device-plugin server stopped serving: %v", err)
				}()

				if !registeredWithKubelet {
					for len(mgm.grpcServer.GetServiceInfo()) <= 0 {
						time.Sleep(1 * time.Second)
					}
					glog.Infoln("device-plugin server started serving")

					err = RegisterToKubelet()
					if err != nil {
						mgm.grpcServer.Stop()
						wg.Wait()
						glog.Fatal(err)
					}
					glog.Infoln("device plugin registered with kubelet")
					registeredWithKubelet = true
				}
			}
		}
	}
}

func RegisterToKubelet() error {
	conn, err := grpc.Dial(kubeletEndpoint, grpc.WithInsecure(),
		grpc.WithDialer(func(addr string, timeout time.Duration) (net.Conn, error) {
			return net.DialTimeout("unix", addr, timeout)
		}))
	if err != nil {
		return fmt.Errorf("device plugin cannot connect to the kubelet service: %v", err)
	}
	defer conn.Close()
	client := pluginapi.NewRegistrationClient(conn)

	request := &pluginapi.RegisterRequest{
		Version:      pluginapi.Version,
		Endpoint:     "mps-device-plugin.sock",
		ResourceName: "nvidia.com/gpu",
	}

	if _, err = client.Register(context.Background(), request); err != nil {
		return fmt.Errorf("device plugin cannot register to kubelet service: %v", err)
	}
	return nil
}

func (ps *pluginService) RegisterService() {
	pluginapi.RegisterDevicePluginServer(ps.mgm.grpcServer, ps)
}

func (mgm *mpsGPUManager) Stop() {
	if mgm.grpcServer != nil {
		mgm.grpcServer.Stop()
	}
	mgm.stop <- true
	<-mgm.stop
	glog.Infof("MPS GPU Manager stopped")
}

func (ps *pluginService) ListDevices() []*pluginapi.Device {
	gpuMemoryAvailable := 16384 // Using static value for now
	computeDevicesCount := 100 / ps.mgm.computePartitionSize
	memoryDevicesCount := gpuMemoryAvailable / ps.mgm.memPartitionSize
	virtualDevices := make([]*pluginapi.Device, computeDevicesCount+memoryDevicesCount)

	for i := 0; i < computeDevicesCount; i++ {
		virtualDeviceID := fmt.Sprintf("%s-%d", computeResourceName, i)
		virtualDevices[i] = &pluginapi.Device{
			ID:     virtualDeviceID,
			Health: pluginapi.Healthy,
		}
	}

	for i := 0; i < memoryDevicesCount; i++ {
		virtualDeviceID := fmt.Sprintf("%s-%d", memResourceName, i)
		virtualDevices[computeDevicesCount+i] = &pluginapi.Device{
			ID:     virtualDeviceID,
			Health: pluginapi.Healthy,
		}
	}
	return virtualDevices
}

func (ps *pluginService) ListAndWatch(empty *pluginapi.Empty, stream pluginapi.DevicePlugin_ListAndWatchServer) error {
	resp := new(pluginapi.ListAndWatchResponse)
	resp.Devices = ps.ListDevices()
	if err := stream.Send(resp); err != nil {
		glog.Infof("Error sending device list : %v", err)
		return err
	}
	glog.Infof("Successfully sent the list of devices ...")
	select {}
}

func (ps *pluginService) Allocate(ctx context.Context, rqt *pluginapi.AllocateRequest) (*pluginapi.AllocateResponse, error) {
	allocateResponse := &pluginapi.AllocateResponse{}

	for _, req := range rqt.ContainerRequests {
		containerAllocateResponse := &pluginapi.ContainerAllocateResponse{}
		totalCompute := len(req.DevicesIDs)
		envs := make(map[string]string)
		envs["CUDA_MPS_ACTIVE_THREAD_PERCENTAGE"] = strconv.Itoa(totalCompute)
		envs["CUDA_MPS_PINNED_DEVICE_MEM_LIMIT"] = "0=2G"
		mounts := []*pluginapi.Mount{
			{
				ContainerPath: "/usr/local/nvidia",
				HostPath:      "/usr/local/nvidia",
			},
		}
		containerAllocateResponse.Envs = envs
		containerAllocateResponse.Mounts = mounts
		allocateResponse.ContainerResponses = append(allocateResponse.ContainerResponses, containerAllocateResponse)
		glog.Infof("Successfully allocated the devices ...")
	}
	return allocateResponse, nil
}

func (ps *pluginService) GetDevicePluginOptions(context.Context, *pluginapi.Empty) (*pluginapi.DevicePluginOptions, error) {
	return &pluginapi.DevicePluginOptions{
		PreStartRequired:                false,
		GetPreferredAllocationAvailable: false,
	}, nil
}

func (ps *pluginService) GetPreferredAllocation(ctx context.Context, rqt *pluginapi.PreferredAllocationRequest) (*pluginapi.PreferredAllocationResponse, error) {
	preferredAllocateResponse := &pluginapi.PreferredAllocationResponse{}

	for _, req := range rqt.ContainerRequests {
		containerAllocateResponse := &pluginapi.ContainerPreferredAllocationResponse{}
		totalDevicesToAllot := req.AllocationSize
		for i := 0; i < len(req.MustIncludeDeviceIDs); i++ {
			containerAllocateResponse.DeviceIDs = append(containerAllocateResponse.DeviceIDs, req.MustIncludeDeviceIDs[i])
			totalDevicesToAllot--
		}
		preferredAllocateResponse.ContainerResponses = append(preferredAllocateResponse.ContainerResponses, containerAllocateResponse)
	}
	return preferredAllocateResponse, nil
}

func (ps *pluginService) PreStartContainer(context.Context, *pluginapi.PreStartContainerRequest) (*pluginapi.PreStartContainerResponse, error) {
	preStartContainerResponse := pluginapi.PreStartContainerResponse{}
	return &preStartContainerResponse, nil
}

func main() {
	mgm := NewMpsGPUManager(1, 256)
	defer mgm.Stop()
	mgm.Serve()
	// if err := mgm.Serve(); err != nil {
	// 	glog.Fatalf("Error starting the MPS GPU Manager : %v", err)
	// }
}
